#!/bin/bash
this_script=$(realpath -s "$0")
scripts_path=$(dirname "$this_script")
real_project_dir="$scripts_path"
PROJECT_DIR="$scripts_path"
startup_params="$*"
startup_params="${startup_params/\$real_project_dir/\%project_dir\%}"

cd $scripts_path

if [ ! -f /.dockerenv ]; then
    git submodule update --init --recursive
    git pull --recurse-submodules

    echo "Starting Docker Container..."
    cd "$(cd "$(dirname "$0")"; pwd)"
    docker compose down
    docker compose pull
    docker compose up -d
    docker exec -it "builder" bash -c "cd /workspace;chmod +x build.sh;./build.sh $startup_params"
    exit $?
fi

#
# Print command (GREEN)
#
# $1 string to print
#
print_command() {
    GREEN='\033[1;32m'
    NC='\033[0m' # No Color
    printf "${GREEN}$1${NC}\n"
}

#
# Print command (GREEN)
#
# $1 string to print
#
print_green() {
    GREEN='\033[1;32m'
    NC='\033[0m' # No Color
    printf "${GREEN}$1${NC}\n"
}

#
# Print blue
#
# $1 string to print
#
print_blue() {
    BLUE='\033[1;34m'
    NC='\033[0m' # No Color
    printf "${BLUE}$1${NC}\n"
}

#
# Print error (RED)
#
# $1 string to print
#
print_red() {
    RED='\033[1;31m'
    NC='\033[0m' # No Color
    printf "${RED}$1${NC}\n"
}

#
# Print error (RED)
#
# $1 string to print
#
print_error() {
    RED='\033[1;31m'
    NC='\033[0m' # No Color
    printf "${RED}$1${NC}\n"
}

#
# Display new section
#
# $1 Section name
#
display_section() {
    print_blue " "
    print_blue "##########################################"
    print_blue "## $*"
    print_blue "##########################################"
    print_blue " "
}

#
# Display error
#
# $1 Error
#
exit_with_error() {
    print_error " "
    print_error "##########################################"
    print_error "## Error: $1"
    print_error "##########################################"
    print_error " "
    if [ ! "$2" ]
    then
        exit 1
    fi
    exit $2
}

export -f print_command
export -f print_red
export -f print_green
export -f print_blue
export -f print_error
export -f display_section
export -f exit_with_error

if [ ! "${CPU_ARCH}" ]
then
  echo CPU_ARCH is not defined, using armhf as default
  CPU_ARCH="armhf"
fi

if [ ! "${CPU_COMPILER_PREFIX}" ]
then
  echo CPU_COMPILER_PREFIX is not defined, using arm-linux-gnueabihf as default
  CPU_COMPILER_PREFIX="arm-linux-gnueabihf"
fi

if [ ! "${MCU_COMPILER_PREFIX}" ]
then
  echo MCU_COMPILER_PREFIX is not defined, using arm-none-eabi as default
  MCU_COMPILER_PREFIX="arm-none-eabi"
fi

if [ ! "${SRC_DIR}" ]
then
  echo SRC_DIR is not defined, using $PROJECT_DIR/txt2-M4-rt-core as default
  SRC_DIR="$PROJECT_DIR/txt2-M4-rt-core"
fi

if [ ! "${BUILD_DIR}" ]
then
  echo BUILD_DIR is not defined, using $PROJECT_DIR/build as default
  BUILD_DIR="$SRC_DIR/build"
fi

if [ ! "${ARTIFACTS_DIR}" ]
then
  echo ARTIFACTS_DIR is not defined, using $PROJECT_DIR/release as default
  ARTIFACTS_DIR="$PROJECT_DIR/release"
fi

if [ ! "${CROSS_COMPILE}" ]
then
  echo CROSS_COMPILE is not defined, using /usr/bin/${CPU_COMPILER_PREFIX}- as default
  CROSS_COMPILE="/usr/bin/${CPU_COMPILER_PREFIX}-"
fi

export LANGUAGE=en_US.UTF-8
export LANG=en_US.UTF-8
export LC_ALL=en_US.UTF-8#
export DEBIAN_FRONTEND=noninteractive
export DEBIAN_PRIORITY=critical
export CPU_ARCH=${CPU_ARCH}
export CPU_COMPILER_PREFIX=${CPU_COMPILER_PREFIX}
export MCU_COMPILER_PREFIX=${MCU_COMPILER_PREFIX}
export BUILD_DIR=${BUILD_DIR}

GLIBC_VERSION="2.28"

prepare_locale()
{
    apt-get update || true
    apt-get upgrade -yq

    apt-get install -yq locales locales-all
    sed -i -e 's/# en_US.UTF-8 UTF-8/en_US.UTF-8 UTF-8/' /etc/locale.gen && \
    dpkg-reconfigure --frontend=noninteractive locales && \
    update-locale LANG=en_US.UTF-8
    ln -fs /usr/share/zoneinfo/Europe/Berlin /etc/localtime || true
}

prepare_repo() 
{
    display_section Preparing APT-Repository...
    . /etc/os-release
    dpkg --add-architecture ${CPU_ARCH}
    echo "deb [arch=amd64,i386] http://us.archive.ubuntu.com/ubuntu/ $UBUNTU_CODENAME main restricted universe multiverse" > /etc/apt/sources.list
    echo "deb [arch=amd64,i386] http://us.archive.ubuntu.com/ubuntu/ $UBUNTU_CODENAME-updates main restricted universe multiverse" >> /etc/apt/sources.list
    echo "deb [arch=amd64,i386] http://us.archive.ubuntu.com/ubuntu/ $UBUNTU_CODENAME-backports main restricted universe multiverse" >> /etc/apt/sources.list
    echo "deb [arch=amd64,i386] http://security.ubuntu.com/ubuntu $UBUNTU_CODENAME-security main restricted universe multiverse" >> /etc/apt/sources.list
    echo "deb [arch=${CPU_ARCH}] http://ports.ubuntu.com/ubuntu-ports/ $UBUNTU_CODENAME main restricted universe multiverse" >> /etc/apt/sources.list
    echo "deb [arch=${CPU_ARCH}] http://ports.ubuntu.com/ubuntu-ports/ $UBUNTU_CODENAME-updates main restricted universe multiverse" >> /etc/apt/sources.list
    echo "deb [arch=${CPU_ARCH}] http://ports.ubuntu.com/ubuntu-ports/ $UBUNTU_CODENAME-backports main restricted universe multiverse" >> /etc/apt/sources.list
    echo "deb [arch=${CPU_ARCH}] http://ports.ubuntu.com/ubuntu-ports/ $UBUNTU_CODENAME-security main restricted universe multiverse" >> /etc/apt/sources.list
    apt update
}

prepare_host_dependencies()
{
    display_section "Preparing host dependencies..."
    apt update
    apt install git swig bison flex cmake automake python3-sip libpython3-dev python3-dev libc6 libc6-dev build-essential libmawk-dev gawk -yq || exit_with_error "Could not install host dependencies"
}

prepare_cross_compiler()
{
    display_section "Preparing cross compiler..."
    apt update
    apt install binutils-${MCU_COMPILER_PREFIX} gcc-${MCU_COMPILER_PREFIX} -yq || exit_with_error "Could not install cross compiler dependencies"
    apt install libc6-dev-armhf-cross libc6-armhf-cross binutils-${CPU_COMPILER_PREFIX} gcc-${CPU_COMPILER_PREFIX} g++-${CPU_COMPILER_PREFIX} pkg-config-${CPU_COMPILER_PREFIX} -yq || exit_with_error "Could not install cross compiler dependencies"
}

prepare_target_dependencies()
{
    display_section "Preparing target dependencies..."
    apt update
    apt install gawk:${CPU_ARCH} libmosquitto-dev:${CPU_ARCH} libmosquittopp-dev:${CPU_ARCH} libssl-dev:${CPU_ARCH} openssl:${CPU_ARCH} libudev-dev:${CPU_ARCH} python3-dev:${CPU_ARCH} libpython3-dev:${CPU_ARCH} -yq || exit_with_error "Could not install target dependencies"

    export CROSS_COMPILE="/usr/bin/${MCU_COMPILER_PREFIX}-"
    if [ -d "/usr/lib/${MCU_COMPILER_PREFIX}" ]
    then
        export LIBDIR="/usr/lib/${MCU_COMPILER_PREFIX}"
    else
        export LIBDIR="/usr/lib"
    fi
    if [ -d "/usr/include/${MCU_COMPILER_PREFIX}" ]
    then
        export INCDIR="/usr/include/${MCU_COMPILER_PREFIX}"
    else
        export LIBDIR="/usr/include"
    fi

    export glibc_install="$(pwd)/glibc/build/install"

    git clone git://sourceware.org/git/glibc.git
    cd glibc
    git checkout glibc-${GLIBC_VERSION}
    mkdir build
    cd build

    #../configure --disable-sanity-checks
    ../configure --prefix "$glibc_install" --disable-werror --enable-cet --host=${CPU_COMPILER_PREFIX}
    make -j $(nproc) >> /dev/null
    make install -j $(nproc)

    ls ${glibc_install}/lib/
    export CFLAGS="${CFLAGS} -L ${glibc_install}/lib -I ${glibc_install}/include -Wl,--rpath=${glibc_install}/lib -Wl,--dynamic-linker=${glibc_install}/lib/ld-linux-armhf.so.3"
    export CXXFLAGS="${CXXFLAGS} -L ${glibc_install}/lib -I ${glibc_install}/include -Wl,--rpath=${glibc_install}/lib -Wl,--dynamic-linker=${glibc_install}/lib/ld-linux-armhf.so.3"

}

cd $PROJECT_DIR

case "${1}" in
    "prepare")
        display_section "Start preparing build..."
        mkdir -p $BUILD_DIR
        mkdir -p $ARTIFACTS_DIR
        rm $BUILD_DIR/* || true
        rm $ARTIFACTS_DIR/* || true
        prepare_locale
        prepare_repo
        prepare_host_dependencies
        prepare_cross_compiler
        prepare_target_dependencies
    ;;
    "build")
        bash $this_script "build_libs" 
        errorcode=$?
        if [ "$errorcode" != "0" ]
        then
            exit_with_error "Building libraries failed..." $errorcode
        fi
        bash $this_script "build_firmware"
        errorcode=$?
        if [ "$errorcode" != "0" ]
        then
            exit_with_error "Building RT core firmware failed..." $errorcode
        fi
    ;;
    "build_libs")
        display_section "Building libraries..."
        export CROSS_COMPILE="/usr/bin/${CPU_COMPILER_PREFIX}-"
        if [ -d "/usr/lib/${CPU_COMPILER_PREFIX}" ]
        then
            export LIBDIR="/usr/lib/${CPU_COMPILER_PREFIX}"
        else
            export LIBDIR="/usr/lib"
        fi
        if [ -d "/usr/include/${CPU_COMPILER_PREFIX}" ]
        then
            export INCDIR="/usr/include/${CPU_COMPILER_PREFIX}"
        else
            export LIBDIR="/usr/include"
        fi
        cd $BUILD_DIR
        PROCESSOR_ARCHITECTURE=arm cmake ../ -DCMAKE_CROSSCOMPILING=1 -DCMAKE_C_COMPILER=${CPU_COMPILER_PREFIX}-gcc -DCMAKE_CXX_COMPILER=${CPU_COMPILER_PREFIX}-g++ -DUSE_PYTHON=YES -DCMAKE_SYSTEM_PROCESSOR=arm -DCMAKE_SYSTEM_NAME=linux
        make
    ;;
    "build_firmware")
        display_section "Building firmware..."
        cd $BUILD_DIR
        export CROSS_COMPILE="/usr/bin/${MCU_COMPILER_PREFIX}-"
        if [ -d "/usr/lib/${MCU_COMPILER_PREFIX}" ]
        then
            export LIBDIR="/usr/lib/${MCU_COMPILER_PREFIX}"
        else
            export LIBDIR="/usr/lib"
        fi
        if [ -d "/usr/include/${MCU_COMPILER_PREFIX}" ]
        then
            export INCDIR="/usr/include/${MCU_COMPILER_PREFIX}"
        else
            export LIBDIR="/usr/include"
        fi
        
        make -f ../Makefile.stm32 clean
        make -f ../Makefile.stm32 all
    ;;
    "after_build")
        ls -l $BUILD_DIR
        rm -R $BUILD_DIR/CMakeFiles || true
        rm $BUILD_DIR/CMakeCache.txt || true
        rm $BUILD_DIR/Makefile || true
        rm $BUILD_DIR/cmake_install.cmake || true
        rm -R $BUILD_DIR/rev0 || true
        rm -R $BUILD_DIR/rev1 || true
        rm $BUILD_DIR/*.map || true
        rm $BUILD_DIR/*.sym || true
        rm $BUILD_DIR/*.bin || true
        cp $SRC_DIR/utils/install.sh $BUILD_DIR/install.sh || true
        chmod +x $BUILD_DIR/install.sh || true
        mkdir -p $BUILD_DIR/includes || true
        cp -drv $SRC_DIR/ft $BUILD_DIR/includes/ || true
        cp -R $BUILD_DIR/* $ARTIFACTS_DIR/
        mkdir -p $ARTIFACTS_DIR/firmware
        mv $ARTIFACTS_DIR/*.elf $ARTIFACTS_DIR/firmware/
        ;;
    "")
        bash $this_script "prepare"
        errorcode=$?
        if [ "$errorcode" != "0" ]
        then
            exit_with_error "Prepare stage failed..." $errorcode
        fi
        bash $this_script "build"
        errorcode=$?
        if [ "$errorcode" != "0" ]
        then
            exit_with_error "Build stage failed..." $errorcode
        fi
        bash $this_script "after_build"
        errorcode=$?
        if [ "$errorcode" != "0" ]
        then
            exit_with_error "After build stage failed..." $errorcode
        fi
        bash $this_script "cleanup"
        errorcode=$?
        if [ "$errorcode" != "0" ]
        then
            exit_with_error "Cleanup stage failed..." $errorcode
        fi
        exit 0;
    ;;
    "cleanup")
        display_section "Cleanup..."
        
        exit 0
    ;;
    *)
        echo "ERROR: not supported..."
    ;;
esac
