/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file TxtControlLib.c
 * @brief Main initialization/communication procedures for user API
 */

#include "../inc/ftconfig.h"
#include "ftlibtypes.h"
#include "ftalloc.h"
#include "fttransport.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/mman.h>

//#define USE_SERVO_CYCLIC_SYNC 1

/**
 @brief Init object structures
 @param txt TXT2 controller
 */
extern void ftop_init_offsets(ftobject_txt2 *txt);

/*
 * Debug mode
 */
int ftdebug = 0; /*!< common debug mask */

/* Must be sycnhronized with fttypes.h */
static char *itypes[] = { "unknown", "switch", "voltage", "resistor",
		"termistor", "color_sensor", "trail_failover", "ultrasonic", "digital",
		"encoder" };

/**
 @brief Get string name for a type
 @param [in] t type
 @return type name
 */
const char *ftmop_get_typename(unsigned int t)
{
	if (t < sizeof(itypes)/sizeof(itypes[0])) {
		return itypes[t];
	} else {
		return "";
	}
}

/**
 @brief Map shared memory structures, must be less than 4K
 @param [in] addr shared memory M4 address
 @return shared memory A7 address
 */
static void *ftopMapValues(void *addr, ftobject_txt2 *txt2)
{
	uint32_t offset = 0x10000000;  /* Physical address */
	int fd = open("/dev/m4mem", O_RDONLY | O_CLOEXEC);
	unsigned char *xaddr;

	if (fd < 0) {
		fd = open("/dev/mem", O_RDONLY | O_CLOEXEC);
		offset = 0;
		if (fd < 0)
			return MAP_FAILED;
	}

	uintptr_t vaddr = (uintptr_t) ((uint32_t)addr - offset);
	uintptr_t paddr = vaddr & (-4096);
	txt2->memid = fd;
	xaddr = mmap(NULL, 4096 * 2, PROT_READ, MAP_SHARED, fd, (off_t) paddr);
	if (xaddr == MAP_FAILED) {
		perror("mmap");
		return xaddr;
	}

	return xaddr + (vaddr - paddr);
}

/**
 @brief Send info request for TXT2
 @param [in] txt TXT2 controller
 */
static void ftmop_get_configuration(ftobject_txt2 *txt)
{
	struct ftop_command_packet get_info = { sizeof(get_info),
				ftop_op_get_info, txt->canid, txt->cmdid++ };

	ftop_socket_send(txt->transport, &get_info, sizeof(get_info));
}

/**
 @brief Debug procedure to output I1-I8 config
 @param [in] iconfig config info
 */
void ftOutputInputsConfig(struct ft_inputs_config *iconfig)
{
    for(int i = 0; i < FT_MAX_INPUTS; i++) {
		fprintf(stderr, "input %d : 0x%x-%s\n", i + 1, iconfig->inputs[i][0],
				itypes[iconfig->inputs[i][1]]);
	}
}

/**
 @brief Debug procedure to output M1-M4 config
 @param [in] mconfig config info
 */
void ftOutputOutputsConfig(struct ft_motors_config *mconfig) {
	for (int i = 0; i < FT_MAX_MOTORS; i++) {
		fprintf(stderr, "output %d : %s\n", i + 1, mconfig->usage[i] ? "motor" : "output");
    }
}

/**
 @brief Debug procedure to output I1-I8 and board values
 @param [in] txt TXT2 controller
 */
static void ShowValues(ftobject_txt2 *txt)
{
    int i;

    fprintf(stderr, "Values\n");
    for(i = 0; i < FT_MAX_MOTORS; i++) {
		fprintf(stderr, "Cmdid[%d] = %ud\n", i,
				txt->current_values.cmdvalues.cmdids[i]);
	}
	for (i = 0; i < FT_MAX_INPUTS; i++) {
		fprintf(stderr, "Input[%d] = 0x%x\n", i, txt->cvalues->inputs[i]);
    }

	fprintf(stderr, "cstate      = %d %d %d %d\n", txt->cvalues->cstate[0],
			txt->cvalues->cstate[1], txt->cvalues->cstate[2],
			txt->cvalues->cstate[3]);
    fprintf(stderr, "temperature = %d C\n", txt->cvalues->temperature);
}

/**
 @brief Copy cmdids values for M1-M4
 @param [in] from initial buffer
 @param [in] to output buffer
 */
static void CopyValues(ftvalues *to, ftcvalues *from)
{
    int i;

    for (i = 0; i < FT_MAX_MOTORS; i++)
	    to->cmdvalues.cmdids[i] = from->cmdids[i];
}

/**
 @brief Process events from M4
 @param [in] txt TXT2 controller
 */
static void ProcessEvents(ftobject_txt2 *txt, uint32_t mask)
{
    /* Produce event */
    pthread_mutex_lock(&txt->emutex);
    txt->events.counts[ft_event_values_update]++;
	while (txt->events.counts[ft_event_slaves_update] < FT_MAX_TXT
			&& txt->suids[txt->events.counts[ft_event_slaves_update]]
					!= 0) {
		txt->events.counts[ft_event_slaves_update]++;
	}
	int ncode = txt->current_values.ft_finfo.faults;
	if (ncode != txt->faultid) {
		txt->events.counts[ft_event_faults_update]++;
		txt->faultid = ncode;
	}
    pthread_cond_broadcast(&txt->econd);
    pthread_mutex_unlock(&txt->emutex);
    if (txt->uaddr == 0) /* Stop stream */
        return;

    struct ftop_set_mode nping = {{ sizeof(nping), ftop_op_nping, 0, txt->cmdid++}, mask};
    ftop_socket_send(txt->transport, &nping, sizeof(nping));
}

/**
 @brief Process input packet from M4
 @param [in] txt TXT2 controller
 */
static void ProcessInputPackets(ftobject_txt2 *txt)
{
    unsigned char buffer[256];
    int code;

    while(1) {
		struct ftop_command_packet *r = (void*) buffer;
		int len = ftop_socket_receive(txt->transport, buffer,
				sizeof(buffer));

		if (len == 0) {
			code = errno;
			perror("receive");
			fprintf(stderr, "Process buffer len %d %d\n", len, code);
			exit(1);
		}

		if (ftdebug & FT_DEBUG_A7_RECEIVE) {
			fprintf(stderr, "Process buffer len %d\n", len);
		}

		uint8_t dst = r->dst;

		/* Select correct TXT2 controller */
		if (dst >= FT_MAX_TXT)
			continue; /* Wrong package */

		ftobject_txt2 *txt2 = txt;
		if (dst) {
			txt2 = txt->slaves[dst - 1].txt2;
			if (txt2 == NULL)
				continue; /* Wrong package */
		}

        /* We ready process this packet from server */
        switch(r->cmd) {
		case ftop_op_ping:
			fprintf(stderr, "Ping\n");
			break;

        case ftop_op_get_info: /* Get TXT2 configuration info */
            {
                struct ftop_response_txt_info *info = (void*)buffer;
                if (ftdebug & FT_DEBUG_A7_INFO) {
					fprintf(stderr, "Read info:\n");
					fprintf(stderr, "version 0x%x\n", info->version);
					fprintf(stderr, "firmware 0x%x\n", info->firmware);
					fprintf(stderr, "uid 0x%8.8X\n", info->uid);
					fprintf(stderr, "mode 0x%x\n", info->mode);
					fprintf(stderr, "slaves 0x%x\n", info->slaves);
					fprintf(stderr, "vaddr 0x%x\n", info->vaddr);
					fprintf(stderr, "taddr 0x%x\n", info->taddr);
					fprintf(stderr, "uaddr 0x%x\n", info->uaddr);
					fprintf(stderr, "name %s\n", info->name);
                }
                txt2->version = info->version;
                txt2->firmware = info->firmware;
                txt2->uid = info->uid;
                txt2->mode = info->mode;
                txt2->nslaves = info->slaves;
                memcpy(txt2->name, info->name, sizeof(txt2->name));
				if (info->uaddr) {
					void *xaddr = ftopMapValues((void*) info->uaddr, txt2);

					if (ftdebug & FT_DEBUG_A7_MAP) {
						fprintf(stderr, "uaddr 0x%x xaddr %p\n", info->uaddr,
								xaddr);
					}

					if (xaddr && xaddr != MAP_FAILED) {
						char *vaddr = (char *) xaddr + (info->vaddr - info->uaddr);
						txt2->uaddr = info->uaddr;
						txt2->header = xaddr;
						txt2->suids = (void*)((uint32_t)txt2->header->uids + (char *)xaddr - info->uaddr);
						txt2->cvalues = (void*) vaddr;
						vaddr = (char *) xaddr + (info->taddr - info->uaddr);
						txt2->txt = (void*) vaddr;
					}
					struct ftop_command_packet get_config = { sizeof(get_config),
							ftop_op_get_config, txt2->canid, txt2->cmdid++ };
					ftop_socket_send(txt2->transport, &get_config,
							sizeof(get_config));
				}
            }
            break;

        case ftop_op_get_config: /* Configuration response */
            {
                struct ftop_response_config *rconfig = (void*)buffer;

                /* Output state */
                if (ftdebug & FT_DEBUG_A7_CONFIG) {
                	fprintf(stderr, "Read config %d\n", rconfig->configid);
                    ftOutputInputsConfig(&rconfig->config.inputs);
                    ftOutputOutputsConfig(&rconfig->config.motors);
                }
            }
            break;

        case ftop_op_get_values: /* Values update */
            {
                struct ftop_response_values *rvalues = (void*)buffer;
                CopyValues(&txt2->current_values, &rvalues->cvalues);
                if (ftdebug & FT_DEBUG_A7_RECEIVE) {
                    fprintf(stderr, "Read values\n");
                    ShowValues(txt2);
                }

                ProcessEvents(txt, 1 << dst);
            }
            break;

        case ftop_op_nresponse:
			{
				struct ftop_nresponse_values *nrvalues = (void*)buffer;
				uint32_t mask = nrvalues->nmask;
				ftcvalues *cvalues = &nrvalues->cvalues[0] - 1;

				for (int i = 0; i < FT_MAX_TXT; i++) {
					if (!(mask & (1 << i)))
						continue;

					cvalues++;
					/* Select TXT */
					ftobject_txt2 *txt2 = i ? txt->slaves[i - 1].txt2 : txt;
					if (txt2 == NULL) {
						continue; /* Wrong package */
					}
					CopyValues(&txt2->current_values, cvalues);
				}

				ProcessEvents(txt, mask);
			}
			break;

        case ftop_op_debug: /* Debug output */
        	fprintf(stderr, "M4 >>> %s", (char*)(r + 1));
        	break;

        default:
            /* Synchroznization problems */
            ftop_socket_free(txt->transport);
            txt->transport = NULL;
            return;
        }
    }
}

/**
 @brief Thread to process input packets from M4
 @param [in] info unused
 */
static void *ftopThread(void *info)
{
    ftobject_txt2 *txt = info;

    ProcessInputPackets(txt);
    return NULL;
}

static int readServoConfigCommand(int num,char* outData,int maxData)
{
	FILE *fp;
	char cmdString[64] = {0};
	sprintf(cmdString,"/usr/bin/sudo -n /usr/sbin/txt_servo getvalues %d",num);
	fp = popen(cmdString, "r");
	if (fp == NULL) {
		return -1;
	}

	while (fgets(outData, maxData, fp) != NULL) {
    	//printf("%s", buffer);
	}
  	pclose(fp);
	return 0;
}

static int parseServoConfigCommand(char* buffer, uint16_t* pu16Min, uint16_t* pu16Max)
{
	char valName[20];
	char* pstr;
	int i;

	memset(valName,0,sizeof(valName));

	pstr=buffer;
	valName[0] = 0;
	while(*pstr != 0)
	{
		while((*pstr != 0) && (*pstr != '"'))
		{
			pstr++;
		}
		if (*pstr == '"') pstr++;
		if (*pstr == 0) return;
		for(i = 0;i < 20;i++)
		{
			if (*(pstr + i) == '"')
			{
				*(pstr + i) = 0;
				if (valName[0] == 0) {
					strncpy(valName,pstr,sizeof(valName));
				} else
				{
					//printf("%s = %s\n",valName,pstr);
					if (strcmp("min",valName) == 0)
					{
						*pu16Min=atoi(pstr);
					} else if (strcmp("max",valName) == 0)
					{
						*pu16Max=atoi(pstr);
					}
					valName[0] = 0;
				}
				pstr += i + 1;
				break;
			}
		}
	}
	return 0;
}

/**
 * @brief Get the Servo Min Max calibration values
 * 
 * @param num Servo number 1-3
 * @param [out] pu16Min Servo minimum calibration value
 * @param [out] pu16Max Servo maximum calibration value
 */
static void getServoMinMax(int num, uint16_t* pu16Min, uint16_t* pu16Max)
{
	char buffer[128];

	memset(buffer,0,sizeof(buffer));

	if ((pu16Min == NULL) || (pu16Max == NULL))
	{
		return;
	}

	*pu16Min = 600;
	*pu16Max = 2400;
	if (readServoConfigCommand(num,buffer,sizeof(buffer)) < 0)
	{
		return;
	}
	parseServoConfigCommand(buffer,pu16Min,pu16Max);
}


static void syncServoCalibration(ftobject_txt2 *txt)
{
	uint16_t u16Min, u16Max;
	void *ftDev = NULL;
	for(int i = 0; i < FT_MAX_SERVO;i++)
	{
		getServoMinMax(i+1,&u16Min,&u16Max);
		if ((txt->dynamic_settings.servo[i].min != u16Min) ||
		(txt->dynamic_settings.servo[i].min != u16Max))
		{
			ftDev = ftop_alloc_servo(txt, ftdev_s1 + i);
			ftop_set_servo_limits(ftDev, u16Min, u16Max);
			ftop_free_device(ftDev);
		}
		txt->dynamic_settings.servo[i].min = u16Min;
		txt->dynamic_settings.servo[i].max = u16Max;
	}
}

#if defined(USE_SERVO_CYCLIC_SYNC)
/**
 @brief Process input packet from M4
 @param [in] txt TXT2 controller
 */
static void ProcessDynamicSettings(ftobject_txt2 *txt)
{
	while(1) {
		syncServoCalibration(txt);
		sleep(30);
	}
}

/**
 @brief Thread to process dynamic settings
 @param [in] info unused
 */
static void *ftsyncDynSettingsThread(void *info)
{
    ftobject_txt2 *txt = info;

    ProcessDynamicSettings(txt);
    return NULL;
}
#endif

/**
 @brief Allocate resources for local TXT
 @param [in] host host with txt controller, "auto" must be used
 @param [in] port connection port (unused)
 @return pointer to TXT2 structure
 */
fttxt2 *ftop_alloc_txt(const char *host, int port)
{
    int err;
    ftop_transport* ft;
    char *debug;
    int timeout = 0;

    if (!strcmp(host, "auto")) /* TODO: extend */
        host = "localhost";

    ft = ftop_socket(host, port);

    if (ft == NULL)
        return NULL;

    debug = getenv("FTDEBUG");
    if (debug != NULL) {
    	sscanf(debug, "%i", &ftdebug);
    }

    ftobject_txt2 *txt = calloc(1, sizeof(txt[0]));
    txt->transport = ft;
    txt->magic = ft_txt2_magic;
    txt->cvalues = &txt->current_values;
    pthread_mutex_init(&txt->emutex, NULL);
    pthread_cond_init(&txt->econd, NULL);
    ftop_init_offsets(txt);

    /* Start thread for communication */
    err = pthread_create(&txt->thread, NULL, ftopThread, txt);
    if (err != 0) {
    	perror("pthread_create");
        free(txt);
        txt = NULL;
    }

	/* Start thread for sync servo calibration settings */
    #if defined(USE_SERVO_CYCLIC_SYNC)
	  err = pthread_create(&txt->sync_settings_thread, NULL, ftsyncDynSettingsThread, txt);
    if (err != 0) {
    	perror("pthread_create");
        free(txt);
        txt = NULL;
    }
	#else
	  syncServoCalibration(txt);
	#endif


	/* Wait and check timeout */
	while (txt->uaddr == 0) {
		/* Get configuration */
		ftmop_get_configuration(txt);
		usleep(10000); /* 10000 mks = 0.01 s */
		timeout++;
		if (timeout > 500) /* 5 sec ! */
			break;
	}

    if (txt->uaddr == 0) {
    	fprintf(stderr, "Connection timeout\n");
    	free(txt);
    	txt = NULL;
    }

    return txt;
}

/**
 @brief Free TXT controller, disconnect and so on
 @param [in] txt txt controller
 */
void ftop_free_txt(fttxt2 *txt)
{
	ftobject_txt2 *txt2 = txt;
	struct ftop_command_packet disconnect = {
		sizeof(disconnect), ftop_op_disconnect, txt2->canid, txt2->cmdid++
    };

	pthread_cancel(txt2->thread);
	#if defined(USE_SERVO_CYCLIC_SYNC)
	pthread_cancel(txt2->sync_settings_thread);
	#endif
	ftop_socket_send(txt2->transport, &disconnect, sizeof(disconnect));
	close(txt2->memid);
        ftop_socket_free(txt2->transport);
}

/**
 @brief Set FDCAN terminator
 @param [in] device TXT2 controller
 @param [in] mode 0 - swith off terminator, 1 switch on terminator
 */
void ftop_set_terminator(void *device, int mode)
{
	ftobject_txt2 *txt = device;
	if (txt->canid)
		return;

	struct ftop_set_mode mcommand = {
			{sizeof(mcommand), ftop_op_terminator, 0, txt->cmdid++}, mode
	};
	ftop_socket_send(txt->transport, &mcommand, sizeof(mcommand));
}

/**
 @brief Set TXT uid
 @param [in] device TXT2 controller
 @param [in] uid - new uid
 */
void ftop_set_uid(void *device, uint32_t uid)
{
	ftobject_txt2 *txt = device;
	if (txt->canid)
		return;

	struct ftop_set_mode mcommand = {
			{sizeof(mcommand), ftop_op_set_uid, 0, txt->cmdid++}, uid
	};
	ftop_socket_send(txt->transport, &mcommand, sizeof(mcommand));
}

/**
 @brief Get uid from shared memory
 @param [in] device master TXT2 controller
 @param [in] number TXT2 number
 @return uid
 */
int ftop_get_specified_uid(void *device, unsigned int number)
{
	ftobject_txt2 *txt = device;

	if (txt->suids == NULL || number >= FT_MAX_TXT)
		return 0;

	return txt->suids[number];
}
