/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @defgroup Common_FT_API_Implementation_Includes
 * @{
 */

/**
 * @file ftext.h
 * @brief Structures, constants and procedures for TXT2 controllers communication
 */

#ifndef FTINC_FTEXT_H_
#define FTINC_FTEXT_H_

#include "ftserver.h"

/**
 * @addtogroup Ext_Communication
 * @{
 */

/*!
 * TXT2 EXT communication commands
 */
enum {
	TXT2_ext_single,  /*!< unused command */
	TXT2_ext_nmaster, /*!< master notification command */
	TXT2_ext_nslave,  /*!< slave notification command */
	TXT2_ext_saddr,   /*!< set address command */
	TXT2_ext_info,    /*!< slave info response */
	TXT2_ext_config,  /*!< slave config command */
	TXT2_ext_op,      /*!< slave operation command */
	TXT2_ext_values   /*!< slave values response */
};

/*!
 * TXT2 EXT common communication packet
 */
typedef struct TXT2_ext_packet {
	ftuid   buid;   /*!< board uid */
	uint8_t action; /*!< TXT2_ext action */
	uint8_t pad1;   /*!< padding */
} TXT2_ext_packet;

/*!
 * TXT2 EXT master notification packet
 */
struct TXT2_ext_master_command {
	ftuid   buid;   /*!< master board uid */
	uint8_t action; /*!< TXT2_ext action */
	uint8_t pad1;   /*!< padding */
	uint32_t tick;  /*!< runtime tick */
};

/*!
 * TXT2 EXT slave notification packet
 */
struct TXT2_ext_slave_command {
	ftuid   buid;   /*!< slave board uid */
	uint8_t action; /*!< TXT2_ext action */
	uint8_t pad1;   /*!< padding */
	uint16_t cid;   /*!< CAN ID if available */
	ftuid   muid;   /*!< master uid if available */
};

/*!
 * TXT2 EXT Set addr packet
 */
struct TXT2_ext_addr_command {
	ftuid   buid;   /*!< master board uid */
	uint8_t action; /*!< TXT2_ext action */
	uint8_t pad1;   /*!< padding */
	uint16_t cid;   /*!< CAN ID */
	ftuid suid;     /*!< slave uid */
};

/*!
 * TXT2 EXT Slave info packet
 */
struct TXT2_ext_info_command {
	ftuid   buid;   /*!< slave board uid */
	uint8_t action; /*!< TXT2_ext action */
	uint8_t pad1;   /*!< padding */
	uint16_t cid;   /*!< CAN ID */
	uint32_t version;               /*!< Slave version */
	union
	{
		uint32_t firmware;              /*!< Slave firmware */
		struct
		{
			uint32_t rtCoreFirmware:16; 
			uint32_t txtFirmware:16;    /* format: decimal xyyzz */
		} firmware_f;
	};
	char device_name[FT_NAME_SIZE]; /*!< Slave TXT2 device name */
	char sn[FT_NAME_SIZE];          /*!< Slave serial number */
};

/*!
 * TXT2 EXT Set config packet for I1-I8 and M1-M4
 */
struct TXT2_ext_config_command {
	ftuid   buid;   /*!< master board uid */
	uint8_t action; /*!< TXT2_ext action */
	uint8_t pad1;   /*!< padding */
	struct ft_remote_config config; /*!< slave TXT2 config */
};

/*!
 * TXT2 EXT operation packet for O1-O8/M1-M4
 */
struct TXT2_ext_op_command {
	ftuid   buid;   /*!< master board uid */
	uint8_t action; /*!< TXT2_ext action */
	uint8_t pad1;   /*!< padding */
	struct ft_remote_op op; /*!< slave TXT2 op */
};

/*!
 * TXT2 EXT slave input values packet
 */
struct TXT2_ext_values_command {
	ftuid   buid;   /*!< slave board uid */
	uint8_t action; /*!< TXT2_ext action */
	uint8_t pad1;   /*!< padding */
	struct ft_remote_values values; /*!< slave TXT2 values */
};

/*!
 * TXT2 EXT packets union
 */
typedef union TXT2_ext_command {
	struct TXT2_ext_packet         pcommand; /*!< EXT common communication packet */
	struct TXT2_ext_master_command mcommand; /*!< EXT master notification packet */
	struct TXT2_ext_slave_command  scommand; /*!< EXT slave notification packet */
	struct TXT2_ext_addr_command   acommand; /*!< EXT Set addr packet */
	struct TXT2_ext_info_command   icommand; /*!< EXT Slave info packet */
	struct TXT2_ext_config_command ccommand; /*!< EXT Set config packet for I1-I8 and M1-M4 */
	struct TXT2_ext_op_command     ocommand; /*!< EXT operation packet for O1-O8/M1-M4 */
	struct TXT2_ext_values_command vcommand; /*!< EXT slave input values packet */
} TXT2_ext_command;

/**
 * @brief Process A7 command for external TXT2 controller
 * @param [in] ftc A7 client
 * @param [in] ftcmd client command
 * @param [in] stxt external TXT2 controller
 */
void ftProcessRequestExtRequest(ftop_client *ftc,
		struct ftop_command_packet *ftcmd, struct ft_shm_txt_remote *stxt);

/**
 * @brief Update config for external TXT2 controller
 * @param [in] ftc A7 client
 * @param [in] stxt external TXT2 controller
 * @param [in] config new config
 * @param [in] dst controller id
 * @param [in] cusage component usage
 */
void UpdateRemoteConfig(ftop_client *ftc, struct ft_shm_txt_remote *stxt, struct ft_config *config, uint32_t dst, uint32_t cusage);

/**
 * @brief Update local configuration for external slave TXT2 controller
 * @param [in] num external TXT2 controller number
 * @return pointer to the configuration copy
 */
struct ft_remote_config *ftExtUpdateConfig(int num);

/**
 * @brief External engine initialization procedure
 */
void ftExtInit();

/** @} */

#endif /* FTINC_FTEXT_H_ */

/** @} */
