/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @defgroup Common_FT_API_Implementation_Includes
 * @{
 */

/**
 * @file ftshmem.h
 * @brief Local shared memory structures
 */

#ifndef FTINC_FTSHMEM_H_
#define FTINC_FTSHMEM_H_

/**
 * @addtogroup Host_Communication_Shared_Memory
 * @{
 */

#include <ftimpltypes.h>

/**
 * Shared memory objects
 */

/**
 * Common TXT2 info
 */
struct ft_shm_common {
    uint32_t version;                /*!< version */
	union
	{
		uint32_t firmware;               /*!< firmware */
		struct
		{
			uint32_t rtCoreFirmware:16; 
			uint32_t txtFirmware:16;    /* format: decimal xyyzz */
		} firmware_f;
	};

    ftuid uid;                       /*!< Processor UID */
    uint32_t mode;                   /*!< ft_txt_mode */
    uint32_t nslave;                 /*!< slave number */
    char device_name[FT_NAME_SIZE];  /*!< Device name */
    char sn[FT_NAME_SIZE];           /*!< Serial number */
};

/**
 * TXT2 shared memory motor object
 */
struct ft_shm_mcommand {
	uint32_t ncmdid;           /*!< New stated motor cmdid */
	uint32_t lcmdid;           /*!< Last executed motor cmdid */
	ft_distance_type distance; /*!< Distance */
};

/**
 * TXT2 shared memory slave state
 */
enum {ftslave_state_none, ftslave_state_addr, ftslave_state_ready};

/**
 * TXT2 shared memory instance object
 */
struct ft_shm_txt {
	struct ft_shm_common common;                    /*!< Common TXT2 part */
	struct ft_config config;                        /*!< Current config */
	struct ftvalues values;                         /*!< Current I1-I8 values */
	ft_counter_type counters[FT_MAX_COUNTERS];      /*!< Current C1-C4 values */
	struct ft_shm_mcommand mcommand[FT_MAX_MOTORS]; /*!< Current motor commands */
	ft_pwm_type     opwm[FT_MAX_OUTPUTS];           /*!< Current O1-O8 PWM values */
	ft_pwm_type     spwm[FT_MAX_SERVO];             /*!< Current S1-S3 PWM values */
    /* Private part */
	int state;                                      /*!< Current TXT2 state */
	uint8_t motormode;								/*!< Motor coast mode */
};

/**
 * TXT2 shared memory statistic information
 */
struct ft_shm_stat {
	uint32_t runtime;			/*!< Runtime from M4 start (ms) */
	uint32_t clients;			/*!< Current active clients set bitmask */
	uint32_t commands;			/*!< number of processes commands from A7 */
	uint32_t adccounter;		/*!< number of ADC1 interrupts (analog inputs) */
	uint32_t spi3counter;		/*!< number of SPI3 interrupts (PWM control) */
	uint32_t usart6counter;		/*!< number of USART6 interrupts (ultrasonic) */
	uint32_t tim7counter;		/*!< number of TIM7 interrupts (digital input and ultrasonic) */
	uint32_t ipccrxcounter;		/*!< number of IPCC rx interrupts */
	uint32_t ipcctxcounter;		/*!< number of IPCC tx interrupts */
	uint32_t maxmotortasktime;	/*!< maximum time between motor task cycles, must be 1 ms */
	uint32_t maxmotortaskcycle; /*!< maximum time in motor task cycle, must be < 1 ms */
	uint32_t maxhwlooptime;		/*!< maximum time in input loop cycle in mks (rev1 only) */
	uint32_t cantx;				/*!< number of TX CAN packets (does not include markers) */
	uint32_t canrx;				/*!< number of RX CAN packets */
	uint32_t pwralert;			/*!< number of PWM alerts (1 ms increment) */
	uint32_t drvalert;			/*!< number of DRV8908 alerts (1 ms increment) */
	uint32_t servoalert;		/*!< number of servo (+5V) alerts (1 ms increment) */
	uint32_t m0maxtime;			/*!< maximum time between PWM0/1 change and M1 count event */
	uint32_t bitrate;			/*!< current CAN bitrate */
	uint32_t fdterm;			/*!< current CAN terminator state */
	uint32_t tdrop;				/*!< number of CAN TX packets dropped by timeout */
};

extern struct ft_shm_stat *ftstat; /*!< Common M4 statistics */
#define FT_SHM_MAGIC "FTSH"			/*!< Magic value for shared memory header */

/**
 * TXT2 shared memory header
 */
struct ft_shm_header {
	char magic[sizeof(uint32_t)]; /*!< Magic value, must be "FTSH" */
	uint32_t mtxt2;               /*!< Maximum TXT2 */
	uint32_t size;                /*!< Total shared memory size */
	ftuid *uids;                  /*!< UIDs in shared memory */
	struct ft_shm_txt **txts;	  /*!< ponters to all TXT controllers, master and slaves */
	struct ft_shm_stat *stat;	  /*!< pointer to statistic information */
	uint32_t *task_times;		  /*!< pointer to tasks statistic information (10 mks resolution) */
};

/**
 * User level visible shared memory
 */
struct ft_shm_userver {
	struct ft_shm_header header;	/*!< header for M4-A7 shared memory */
};

#endif /* FTINC_FTSHMEM_H_ */

/** @} */

/** @} */
