#include <unistd.h>
#include <sys/file.h>
#include <string.h>
#include <errno.h>
#include <sys/inotify.h>
#include <sys/types.h>

#include "ftlock.h"

#define FTLOCK_FILE "/var/tmp/ftlock" /* ftlock file */
#define FTLOCK_MAX_NAME 128		/* maximum name */

static int lid = -1;  /* File descriptor for lock */
static int tfd = -1;  /* File descriptor for inotify */
static int code = -1; /* Last notification */
static int mode = -1; /* Lock state */

/*
   Lock file structure
*/
static struct record {
    int status;
    pid_t pid;
    char name[FTLOCK_MAX_NAME + 1];
} record;

int ftlock_open(void)
{
    if (lid < 0) {
        lid = open(FTLOCK_FILE, O_RDWR | O_CREAT | O_CLOEXEC, 0666);
    }
    return lid;
}

static int lock(char *name, int lck)
{
    if (lid < 0)
        return -1;

    struct flock lock = {.l_type = F_WRLCK, .l_whence = SEEK_SET, .l_len = sizeof(pid_t) + FTLOCK_MAX_NAME + 1, .l_start = sizeof(int)};
    mode = fcntl(lid, lck, &lock);
    if (mode)
        return mode;

    if (name == NULL) name = "";
    strncpy(record.name, name, sizeof(record.name) - 1);
    record.status = 0;
    record.pid = getpid();

    if (lseek(lid, 0, SEEK_SET) < 0)
        return -1;

    if (write(lid, &record, sizeof(record)) != sizeof(record))
        return -1;

    lock.l_start = 0;
    lock.l_len += sizeof(int);
    return fcntl(lid, F_SETLK, &lock);
}

int ftlock_lock(char *name)
{
    return lock(name, F_SETLKW);
}

int ftlock_try_lock(char *name)
{
    return lock(name, F_SETLK);
}

int ftlock_notify(int code)
{
    if (mode < 0)
        return mode;

    if (lseek(lid, 0, SEEK_SET) < 0)
        return -1;
    
    if (write(lid, &code, sizeof(code)) != sizeof(code))
        return -1;

    return 0;
}

/* Space for 128 + 1 char is required */
int ftlock_get_state(char *name, int *pid)
{
    struct record r;
    int res;

    if (name != NULL) {
        name[0] = 0;
        name[FTLOCK_MAX_NAME] = 0;
    }

    if (lid < 0)
        return -1;

    struct flock lock = {.l_type = F_WRLCK, .l_whence = SEEK_SET, .l_len = 1, .l_start = 0, .l_pid = -1};
    res = fcntl(lid, F_GETLK, &lock);
    if (res < 0)
        return res;

    if (lseek(lid, 0, SEEK_SET) < 0)
        return -1;
    
    if (read(lid, &r, sizeof(r)) != sizeof(r))
        return -1;

    res = r.status;

    if (lock.l_pid == -1) {
        lock.l_type = F_WRLCK;
        lock.l_start = sizeof(int);
        if (fcntl(lid, F_GETLK, &lock) < 0 || lock.l_pid != -1)
            return -1; // Something is trying to acquire the lock -> record is invalid
    } else if (lock.l_pid != r.pid)
        return -1;

    if (name != NULL)
        strncpy(name, r.name, FTLOCK_MAX_NAME);

    if (pid != NULL)
        *pid = r.pid;

    return res;
}

ftlock_status ftlock_state()
{
    ftlock_status s;
    s.status = ftlock_get_state(s.name, &s.pid);
    return s;
}

/* Can not return pid_t as swig does not like it */
int ftlock_pid(void)
{
    if (lid < 0)
        return -1;
    struct flock lock = {.l_type = F_WRLCK, .l_whence = SEEK_SET, .l_len = 1, .l_start = sizeof(int), .l_pid = -1};
    fcntl(lid, F_GETLK, &lock);
    return lock.l_pid;
}

int ftlock_unlock(void)
{
    if (mode < 0)
        return mode;
    struct flock lock = {.l_type = F_UNLCK, .l_whence = SEEK_SET, .l_len = sizeof(record), .l_start = 0};
    return fcntl(lid, F_SETLK, &lock);
}

int ftlock_close(void)
{
    mode = -1;
    int res = close(lid);
    lid = -1;
    return res;
}

static int ftlock_inotify_init(void)
{
    tfd = inotify_init();
    return(inotify_add_watch(tfd, FTLOCK_FILE, IN_MODIFY));
}

int ftlock_wait(void)
{
    if (lid < 0)
        return -1;

    if (tfd < 0) {
        if (ftlock_inotify_init() < 0)
            return -1;
        code = -1;
    }

    while (lid >= 0) {
        int len;
        char buf[256] __attribute__ ((aligned(__alignof__(struct inotify_event))));
        int ic = ftlock_get_state(NULL, NULL);
      
        if (ic < 0 || ic != code) {
            code = ic;
            return ic;
        }

        len = read(tfd, buf, sizeof buf);
        if (len < 0 && errno != EAGAIN)
            return -1;
    }

    return -1;
}

