/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/*
 * @file ftouttask.c
 * @brief Notification task
 */

#include "ftconfig.h"
#include "ftservermem.h"
#include "ftosdep.h"
#include "main.h"
#include "ftcalib.h"
#include "ftserver.h"

#include <FreeRTOS.h>
#include <cmsis_os.h>
#include <semphr.h>

/* Total response for all slaves */
static struct ftop_nresponse_values nrvalues = {
    {sizeof(nrvalues), ftop_op_nresponse, 0, 0},
};

extern ftop_client clients[FT_MAX_CLIENTS]; /*!< information about local clients connection */
extern OS_MUTEX(out_mutex);

/**
 * @brief Send TXT2 controller values to A7 client
 * @param dst TXT2 controller number
 * @param ftc A7 client
 */
static void SendValues(int dst, ftcvalues *cvalues, ftop_client *ftc)
{
	/* Calculate new cmdid values for this client */
	struct ftop_motor_for_client *mclient = ftc->motor_cmdid[dst];
	struct ft_shm_mcommand *mcommand = txt2_server.txts[dst]->mcommand;

	OS_LOCK(out_mutex);
	for (int i = 0; i < FT_MAX_MOTORS; mcommand++, mclient++, i++) {
		if (mclient->lcmdid != mclient->ncmdid) {
			int32_t diff = mcommand->lcmdid - mclient->mcmdid;
			if (diff >= 0)
				mclient->lcmdid = mclient->ncmdid;
		}
		cvalues->cmdids[i] = mclient->lcmdid;
	}
	OS_UNLOCK(out_mutex);
}

/**
 * @brief The thread to send client specific information, for example, motor states
 * @param info unused
 */
void sendThread(void const *info) {
	(void) info;

	while (1) {
		extern struct ftcalib ftcalib;
		uint32_t ctime = HAL_GetTick();

		/* Send response to all active clients */
		for (int i = 0; i < FT_MAX_CLIENTS; i++) {
			ftop_client *ftc = &clients[i];
			if ((ctime - ftc->ltick) < ftcalib.rt)
				continue;

			ftcvalues *cvalues = &nrvalues.cvalues[0];
			nrvalues.packet.len = sizeof(nrvalues.packet) + sizeof(nrvalues.nmask);
			nrvalues.nmask = 0;

			for (int j = 0; j < FT_MAX_TXT; j++) {
				if (ftc->opened[j]) {
					ftc->opened[j] = 0; /* Ping is required for active client */

					/* Also send values */
					SendValues(j, cvalues, ftc);
					nrvalues.nmask |= 1 << j;
					nrvalues.packet.len += sizeof(cvalues[0]);
					cvalues++;
				}
			}
			if (nrvalues.nmask) {
				ftserver_send_response(ftc, &nrvalues, nrvalues.packet.len);
				ftc->ltick = ctime;
			}
		}
		osDelay(1);
	}
}


