/*
 * @file rpmsgServer.c
 * @brief RPMSG transport file for M4 to send/receive message between M4 and A7
 *
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "fttransport.h"
#include "ftserver.h"

/* OpenAMP */
#include "openamp.h"

#include <FreeRTOS.h>
#include <semphr.h>

/* Private define */

#define RPMSG_SERVICE_NAME   "rpmsg-raw-channel"

/* Private variables */

static struct rpmsg_endpoint rpmsg_server_ept;
static StaticSemaphore_t hwCommandSemaphore;

/* External variables */

SemaphoreHandle_t commandHandle; /*!< RTOS handle for command transport interface interrupt */
extern ftop_client clients[FT_MAX_CLIENTS];

/* Private functions*/

/**
 * @brief Callback function for reception on virtual channel 0
 * @param ept endpoint
 * @param data buffer
 * @param len buffer size
 * @param src remote endpoint address
 * @param priv unused
 */
static int Server_read_cb(struct rpmsg_endpoint *ept, void *data, size_t len,
		uint32_t src, void *priv)
{
	unsigned int client = src - 2;

	if (client >= FT_MAX_CLIENTS)
		return RPMSG_ERR_ADDR;

	/* Process command from A7 */
	ftProcessRequest(&clients[client], (unsigned char *) data);
	return RPMSG_SUCCESS;
}

/* Exported functions */

/**
 * @brief Rpmsg command server initialization procedure
 */
void Init_Rpmsg_Server(void)
{
	/* Create semaphore */
	commandHandle = xSemaphoreCreateBinaryStatic(&hwCommandSemaphore);

	/* Create a endpoint for rmpsg communication */
	OPENAMP_create_endpoint(&rpmsg_server_ept, RPMSG_SERVICE_NAME, RPMSG_ADDR_ANY,
			  Server_read_cb, NULL);
}

/**
 * @brief Send RPMSG response to A7 client, wakeup sendThread if required
 * @param ftc client
 * @param buffer buffer with response
 * @param len length of information
 * @return length of information
 */
int ftserver_send_response(ftop_client *ftc, void *buffer, int len)
{
	while (rpmsg_send_offchannel_raw(&rpmsg_server_ept, 0, ftc->socket, buffer, len, 1) != len) {
		taskYIELD();
	}
	return len;
}

/**
 * @brief Command server task to process commands from A7 clients
 * @param argument unused
 */
void appmain(void const * argument)
{
	(void) argument;
	extern struct rpmsg_virtio_device rvdev;

	/* Main loop for commands, real message is processed in Server_read_cb */
	while(1) {
		if (MAILBOX_Poll(rvdev.vdev) >= 0)
			continue;

		xSemaphoreTake(commandHandle, 1);
	}
}
