/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ftdaemon.c
 * @brief Demon to stop firmware activity after client abort
 */

#include "fttransport.h"
#include "ftlibtypes.h"
#include "ftalloc.h"

#include <poll.h>
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#define __USE_XOPEN_EXTENDED
#include <signal.h>
#include <sys/inotify.h>
#include <sys/sysmacros.h>

//static FILE *ft;

static int verbose;
static char filename[] = "/dev/rpmsg0";
static struct ftobject_txt2 *txt;
static struct inotify_event event;
static int fd;
static int lid;

static void *checkThread(void *info)
{
	(void)info;

	int tfd = inotify_init();
	int wd = inotify_add_watch(tfd, "/dev/rpmsg_ctrl0", IN_DELETE_SELF);
    if (wd < 0) {
        perror("inotify_add_watch");
        exit(1);
    }
    read(tfd, &event, sizeof(event));
    fprintf(stderr, "No connection\n");
    exit(2);
	return NULL;
}

static void CheckClient(int n)
{
	if (ftop_lock_client(lid, n) == -1) {
		/* OK, busy */
		return;
	}

	/* Try to open client file /dev/rpmsg */
	filename[10] = '0' + n;
	int id = open(filename, O_RDWR | O_EXCL);
	if (id < 0) {
		ftop_unlock_client(lid, n);
		return; /* OK */
	}

    /* HACK: Send free op */
    struct ftop_command_packet disconnect = {
		sizeof(disconnect), ftop_op_disconnect, 0, 0
    };

    write(id, &disconnect, sizeof(disconnect));

    /* Clean buffers */
    fcntl(id, F_SETFL, O_RDWR | O_EXCL | O_NONBLOCK);

    char buffer[128];
    while (read(id, buffer, sizeof(buffer)) > 0);

    close(id);

    ftop_unlock_client(lid, n);
    //fprintf(ft, "Free %d\n", n);
}

static void CheckClients(int mask)
{
	int n;

	for (n = 0; n < FT_MAX_CLIENTS; n++) {
		if (mask & (1 << n))
			CheckClient(n);
	}
}

/**
 * @brief Standard main procedure
 * @param [in] argc number of arguments
 * @param [in] argv arguments
 * @return Error code
 */
int main(int argc, char **argv)
{
    pthread_t thread;
    struct pollfd fds[1];
    //ft = fopen("/dev/tty", "a");
    //setlinebuf(ft);

	verbose = (argc > 1 ) && !strcmp(argv[1], "-v");

	if (!verbose)
		daemon(1, 0);

	sigignore(SIGTERM);

	txt = ftop_alloc_txt("auto", 65000);

	if (txt == NULL) {
		if (verbose)
			fprintf(stderr, "Can't connect to TXT2\n");
		exit(1);
	}

	pthread_cancel(txt->thread);

	/* Get statistic memory */
	struct ft_shm_header *header = txt->header;
	struct ft_shm_stat volatile *stat
		= (struct ft_shm_stat *)((char *)header + ((uint32_t)header->stat - txt->uaddr));

	fd = txt->transport->fid;
	lid = txt->transport->lfid;
	int mdev = txt->transport->cid;
	int tfd = inotify_init();

	if (verbose)
		printf("id %d\n", fd);

	if (verbose)
		printf("client id %d\n", mdev);

	/* Start thread to check status */
	int err = pthread_create(&thread, NULL, checkThread, txt);
	if (err != 0) {
		perror("pthread_create");
		free(txt);
		txt = NULL;
		exit(1);
	}

	int mmask = 1 << mdev;

	/* prepare inotify call */
	int wd = inotify_add_watch(tfd, "/run/lock/LCK..ftrpmsg", IN_CLOSE);
	if (wd < 0) {
		perror("inotify_add_watch");
		exit(1);
	}

	fds[0].fd = tfd;
	fds[0].events = POLLIN;

	while (1) {
		static struct inotify_event event;
		int cmask = stat->clients & ~mmask;

		int err = poll(fds, 1, 1000); /* 1 sec timeout */

		if (err < 0 && errno == EINTR) {
			//fprintf(ft, "eintr\n");
			continue;
		}

		if (err > 0) {
			int size = read(tfd, &event, sizeof(event));
			(void)size;
			//fprintf(ft, "revent %d\n", size);
			cmask = ((1 << FT_MAX_CLIENTS) - 1) & ~mmask;
		} else {
			;//fprintf(ft, "timeout\n");
		}

		//fprintf(ft, "cmask 0x%x\n", cmask);

		if (cmask) {
			if (verbose)
				printf("cmask 0x%x\n", cmask);
			CheckClients(cmask);
		}
	}

    return 0;
}
