/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file HWInit.h
 * @brief Some common rev0/rev1 initialization algorithms
 */

#include "main.h"
#include "HWFDCAN.h"
#include "ftcalib.h"
#include "ftext.h"
#include "HWServo.h"
#include "DRV8908.h"

#include "cmsis_os.h"

/* Private variables */

static uint32_t Prescaler;
static uint32_t PrescalerI18;
static uint32_t MuxerDelay;
static uint32_t spi3divider;
static uint32_t spi3idleness;
static uint32_t adcsampling;

/* External variables */

extern struct ftcalib ftcalib;

ADC_HandleTypeDef hadc1;
IPCC_HandleTypeDef hipcc;

/* External functions */

void DRV8908_init();
void HWLoop_Init();
void HWCountersInit();

/* Private functions */

static void HWInit()
{
	uint32_t pll_freq;
	int j;

	/* TIMG1 - TIM2-TIM7, TIM12-TIM14 */
	Prescaler = (HAL_RCCEx_GetPeriphCLKFreq(RCC_PERIPHCLK_TIMG1) / 1000000) - 1; /* 1 mcs */
	PrescalerI18 = (HAL_RCCEx_GetPeriphCLKFreq(RCC_PERIPHCLK_TIMG1) / 1000000
			/ (1000 / I1I8TimeResolution)) - 1; /* 200 ns*/
	MuxerDelay = (I1I8Delay / I1I8TimeResolution) + 1; /* Check and fix */

	pll_freq = HAL_RCCEx_GetPeriphCLKFreq(RCC_PERIPHCLK_SPI23);

	int i = 2; /* Minimum SPI prescaler */
	spi3divider = SPI_BAUDRATEPRESCALER_2;
	while (i * 5000000 < pll_freq) { /* DRV8908 requires frequency <= 5 Mhz */
		i *= 2; /* Only power of 2 can be used for SPI prescaler */
		spi3divider += SPI_BAUDRATEPRESCALER_4;
	}
	/* DRV8908 write gap >= 2.5 mks */
	i = 1000000000 / pll_freq * i; /* cycle time */
	for (j = 1; j * i < 2500; j++)
		;

	spi3idleness = SPI_MASTER_INTERDATA_IDLENESS_01CYCLE * j;

	{ /* calibration code */
		uint32_t uid = buid = 0x12345678;
		if (uid == ftcalib.uid) {
			adcsampling = ftcalib.adcs & 7;
		} else {
			adcsampling = ADC_SAMPLETIME_32CYCLES_5;
		}
		ftcalib.uid = uid;
		ftcalib.mmark = uid;
		for (int k = 1; k < FT_MAX_INPUTS; k++) {
			struct ftcalib_group *fg = &ftcalib.channel[k];
			if (!fg->div1)
				fg->div1 = ftcalib.channel[0].div1;
			if (!fg->div2)
				fg->div2 = ftcalib.channel[0].div2;
			if (!fg->val1)
				fg->val1 = ftcalib.channel[0].val1;
			if (!fg->val2)
				fg->val2 = ftcalib.channel[0].val2;
		}
	}
}

static void SWInit()
{
	ADC_MultiModeTypeDef multimode = { 0 };
	if (HAL_ADCEx_MultiModeConfigChannel(&hadc1, &multimode) != HAL_OK) {
		Error_Handler();
	}

	while (HAL_ADCEx_Calibration_Start(&hadc1, ADC_CALIB_OFFSET_LINEARITY,
			ADC_SINGLE_ENDED) != HAL_OK)
		;

	/* Init C1-C4 counters */
	HWCountersInit();
	/* Init DRV8908 to control O1-O8 */
	DRV8908_init();
	/* Init main input loop for I1-I8 */
	HWLoop_Init();
	/* Init servo motors for S1-S3 */
	HWServo_Init();
	/* Init FDCAN */
	HWFDCAN_init();
	/* Init External TXT2 controllers engine */
	ftExtInit();
#if COLLECT_TASK_INFO
	TIM5->CR1 |= TIM_CR1_CEN;
#endif
}

/**
 * @brief PowerDown procedure to free resources and stop any activity
 */
void PowerDown()
{
    vTaskSuspendAll ();
    /* Stop DRV8908 and Remove 9V */
    DRV8908_Stop();
    HWPowerDown();
}

/**
 * @brief Procedure to shutdown M4 for firmware "stop" command
 */
static void ShutdownCallback(IPCC_HandleTypeDef * hipcc, uint32_t ChannelIndex,
		IPCC_CHANNELDirTypeDef ChannelDir)
{
	log_info("shutdown");
	PowerDown();
}

static void ThreadsInit()
{
	Init_Rpmsg_Server();
	log_info("After Init_Rpmsg_Server\n");

	/* Setup powerdown callback for M4 "stop" command */
	HAL_IPCC_ActivateNotification(&hipcc, IPCC_CHANNEL_3, IPCC_CHANNEL_DIR_RX,
			ShutdownCallback);

	ftInitServer();
	log_info("After ftInitServer\n");
}
