/*
 * Copyright (c) 2020 RTSoft. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * @file ExtEngine.c
 * @brief External Engine TXT2 communication procedures
 */
#include "main.h"
#include "ExtEngine.h"
#include "HWFDCAN.h"
#include "ftcalib.h"
#include "ftservermem.h"

#include "cmsis_os.h"
#include <string.h>

extern struct ftcalib ftcalib; /*!< Common calibration values */
int nslaves; /*!< Number of slaves at this moment */

/*
 * Slave state
 */
enum {ftsstate_unknown, ftsstate_send_info, ftsstate_send_values};

static int slaveState;

static int canAddr;
static ftuid master_uid;
static uint32_t master_configid;
static struct ft_master_remote master_remote;

uint32_t master_usage = 0; /*!< bitmask for devices used by master */

extern uint32_t slavePacketTime[FT_MAX_TXT * 2 + 4]; /*!< Timeout for errata */

/**
 * @brief Search/add shared memory structure for the slave controller by uid
 * @param uid controller uid
 * @return shared memory structure pointer for the slave controller
 */
static struct ft_shm_txt *Ext_TXT2_Add(ftuid uid)
{
	struct ft_shm_txt *txt;
    int i;

	for (i = 0; i < FT_MAX_TXT; i++) {
		txt = txt2_server.txts[i];
		if (txt2_server.uids[i] == uid){
			return txt;
		}
		if (!txt->common.uid) { /* First free slot */
			break;
		}
	}

	if (i >= FT_MAX_TXT)
		return NULL;

    /* Add new element */
    txt->common.uid = uid;
    txt->common.nslave = nslaves = i;
    txt->common.mode = ft_txt_slave;
    txt2_server.uids[i] = txt->common.uid;
    txt->state = ftslave_state_addr;

    return txt;
}

/**
 * @brief Process new values from remote TXT2 controller
 * @param txt remote controller structure
 * @param values new values
 */
static void ProcessNewValues(struct ft_shm_txt *txt, struct ft_remote_values *values)
{
	struct ft_shm_txt_remote *r = (void*)txt;
	int i;

	if (values->valuesid == r->valuesid)
		return;

	for (i = 0; i < FT_MAX_INPUTS; i++) {
		r->common.values.inputs[i] = values->values.inputs[i];
	}

	for (i = 0; i < FT_MAX_MOTORS; i++) {
		r->common.mcommand[i].lcmdid = values->values.cmdvalues.cmdids[i];
	}

	for (i = 0; i < FT_MAX_COUNTERS; i++) {
		r->common.values.counters[i] = values->values.counters[i];
	}

	*(uint32_t *)r->common.values.cstate = *(uint32_t *)values->values.cstate;
	r->common.values.ft_finfo = values->values.ft_finfo;
}

/**
 * @brief Check buffer pending bit
 * @param nbuf [in] buffer number
 */
static int HWFDCAN_buffer_busy(int nbuf)
{
	if ((FDCAN1->TXBRP & (1 << nbuf)) != 0) {
		if ((HAL_GetTick() - slavePacketTime[nbuf]) > ftcalib.ct) {
			/* Cancel request */
			FDCAN1->TXBCR = 1 << nbuf;
			ftstat->tdrop++;
		}
		return 1;
	}
	return 0;
}

/**
 * @brief Prepare new package for master (discovery, info, values)
 * @param [in] ecommand buffer for new package
 */
void ProcessSlaveTxPacket(TXT2_ext_command *ecommand)
{
	uint32_t uid = buid;
	uint32_t uid16 = uid + (uid >> 16);

	switch(slaveState) {
	case ftsstate_unknown:
		if (!HWFDCAN_buffer_busy(FT_CAN_TXN_DISCOVERY)) {
			struct TXT2_ext_slave_command *command = &ecommand->scommand;

			/* Send discovery message */
			command->action = TXT2_ext_nslave;
			command->buid = buid;
			command->cid = canAddr;
			command->muid = master_uid;
			uint32_t taddr = (HAL_GetTick() ^ uid16) & ((1 << 29) - 1);
			if ((taddr & -(1 << 11)) == 0) {
				taddr = ~taddr & ((1 << 29) - 1);
			}
			HWFDCAN_set_buffer((uint8_t *) ecommand, sizeof(command[0]), taddr,
					FT_CAN_TXN_DISCOVERY);
		}
		break;
	case ftsstate_send_info:
		if (!HWFDCAN_buffer_busy(FT_CAN_TXN_INFO(canAddr))) {
			struct TXT2_ext_info_command *command = &ecommand->icommand;

			command->action = TXT2_ext_info;
			command->buid = buid;
			struct ft_shm_txt *txt = getTXT(0);
			command->firmware = txt->common.firmware;
			command->version = txt->common.version;
			command->cid = canAddr;
			memcpy(command->device_name, txt->common.device_name,
					sizeof(command->device_name));
			memcpy(command->sn, txt->common.sn, sizeof(command->sn));
			HWFDCAN_set_buffer((uint8_t *) ecommand, sizeof(command[0]),
					FT_CAN_SLAVE_TX_ADDR(canAddr), FT_CAN_TXN_INFO(canAddr));
		}
		break;
	case ftsstate_send_values:
		if (!HWFDCAN_buffer_busy(FT_CAN_TXN_VALUES(canAddr))) {
			struct TXT2_ext_values_command *command = &ecommand->vcommand;

			command->action = TXT2_ext_values;
			command->buid = buid;
			getValues(&command->values);

			HWFDCAN_set_buffer((uint8_t *) ecommand, sizeof(command[0]),
					FT_CAN_SLAVE_TX_ADDR(canAddr), FT_CAN_TXN_VALUES(canAddr));
		}
		break;
	}

}

/**
 * @brief Process new package from master (set addr, config, op)
 * @param [in] CanRxData buffer with new package
 */
void ProcessPacketFromMaster(void *CanRxData)
{
	TXT2_ext_packet *p = CanRxData;

	switch(p->action) {
	case TXT2_ext_saddr:
		{
			struct TXT2_ext_addr_command *a = (void*)p;
			if (a->suid != buid)
				break;

			master_uid = a->buid;
			canAddr = a->cid;
			HWFDCAN_AddAddressToFilter(canAddr);
			slaveState = ftsstate_send_info;
			master_configid = 0;
			/* TODO: clean config/op */
		}
		break;

	case TXT2_ext_config:
		{
			struct TXT2_ext_config_command *c = (void*)p;

			if (c->buid != master_uid)
				break; /* Something wrong */

			/* Some optimization */
			if (master_configid == c->config.configid)
				break;

			master_configid = c->config.configid;
			master_usage = c->config.usage;
			ProcessNewConfig(c->config.usage, &c->config.config);
			slaveState = ftsstate_send_values;
		}
		break;

	case TXT2_ext_op:
		{
			struct TXT2_ext_op_command  *c = (void *)p;
			if (c->buid != master_uid)
				break; /* Something wrong */

			ProcessRemoteOp(&master_remote, &c->op);
			slaveState = ftsstate_send_values;
		}
		break;
	}
}

/**
 * @brief Process discovery package from slave
 * @param [in] CanRxData buffer with new package
 */
void ProcessDiscoveryFromSlave(void *CanRxData)
{
    /* 	CanRxData contains discovery package */
	TXT2_ext_packet *command = CanRxData;

	if (masterSilentMode)
		return;

	switch(command->action) {
	case TXT2_ext_nslave: /* Try to select address */
		{
			struct TXT2_ext_slave_command *sc = (void *)CanRxData;
			if (sc->muid && sc->muid != buid) {
				break; /* TODO: Looks like a conflict */
			}

			struct ft_shm_txt *stxt = Ext_TXT2_Add(sc->buid);
			if (stxt && stxt->common.nslave != sc->cid) {
				stxt->state = ftslave_state_addr;
			}
		}
		break;

	default:
		/* Someting wrong */
		; //log_info("Illegal slave package %d\n", command->action);
	}
}

/**
 * @brief Process package from slave or another master (master/slave info, values)
 * @param [in] CanRxHeader can header
 * @param [in] CanRxData buffer with new package
 */
void ProcessPacketFromSlave(FDCAN_RxHeaderTypeDef *CanRxHeader,void *CanRxData)
{
	TXT2_ext_packet *p = CanRxData;

	switch(p->action) {
	case TXT2_ext_nmaster:
		{
			struct TXT2_ext_master_command *mcommand = (void*) CanRxData;
			if (mcommand->buid == buid) /* Is it possible ? */
				break;
			uint32_t tick = HAL_GetTick() - master_tick;
			if (tick > mcommand->tick)
				break;
			masterSilentMode = 1;
			masterSilentTick = HAL_GetTick();
		}
		break;

	case TXT2_ext_info:
		{
			struct TXT2_ext_info_command *icommand = (void*) CanRxData;
			uint16_t n = icommand->cid;
			if (n <= 0 || n > FT_MAX_SLAVES || masterSilentMode)
				break;

			struct ft_shm_txt *txt = getTXT(n);
			if (txt->common.nslave != n || txt->common.uid != icommand->buid) {
				struct ft_shm_txt *stxt = Ext_TXT2_Add(icommand->buid);
				stxt->state = ftslave_state_addr;
				break;
			}
			txt->common.firmware = icommand->firmware;
			txt->common.version = icommand->version;
			memcpy(txt->common.device_name, icommand->device_name,
					sizeof(txt->common.device_name));
			memcpy(txt->common.sn, icommand->sn, sizeof(txt->common.sn));
			txt->state = ftslave_state_ready;
		}
		break;

	case TXT2_ext_values:
		{
			struct TXT2_ext_values_command *vcommand = (void*) CanRxData;
			uint16_t n = (CanRxHeader->Identifier - FT_CAN_BASE_ID - 1) / 2;
			if (n <= 0 || n > FT_MAX_SLAVES || masterSilentMode)
				break;

			struct ft_shm_txt *txt = getTXT(n);
			if (txt->common.nslave != n || txt->common.uid != vcommand->buid) {
				struct ft_shm_txt *stxt = Ext_TXT2_Add(vcommand->buid);
				stxt->state = ftslave_state_addr;
				break;
			}
			ProcessNewValues(txt, &vcommand->values);
		}
		break;
	}
}

/**
 * @brief Send CAN ID from master to slave
 * @param txt slave controller
 * @param ecommand command buffer
 */
void SendAddrPacket(struct ft_shm_txt *txt, TXT2_ext_command *ecommand)
{
	struct TXT2_ext_addr_command *command = &ecommand->acommand;
	command->action = TXT2_ext_saddr;
	command->buid = buid;
	command->suid = txt->common.uid;
	command->cid = txt->common.nslave;
	HWFDCAN_set_buffer((uint8_t *) ecommand, sizeof(*command),
						FT_CAN_DISCOVERY_ID, FT_CAN_TXN_SADDR(txt->common.nslave));
}

/**
 * @brief Send new config from master to slave
 * @param nslave slave controller number
 * @param ecommand command buffer
 */
void SendConfigPacket(int nslave, TXT2_ext_command *ecommand)
{
	struct ft_remote_config *rconfig = ftExtUpdateConfig(nslave);
	struct TXT2_ext_config_command *command = &ecommand->ccommand;
	command->buid = buid;
	command->config = *rconfig;
	command->action = TXT2_ext_config;
	HWFDCAN_set_buffer((uint8_t *) ecommand, sizeof(*command),
			FT_CAN_MASTER_TX_ADDR(nslave), FT_CAN_TXN_CONFIG(nslave));
}

/**
 * @brief Send new operation for O1-O8 from master to slave
 * @param txt slave controller
 * @param ecommand command buffer
 */
void SendOpPacket(struct ft_shm_txt *txt, TXT2_ext_command *ecommand)
{
	struct TXT2_ext_op_command *command = &ecommand->ocommand;
	command->buid = buid;
	command->action = TXT2_ext_op;
	getOp((struct ft_shm_txt_remote *)txt, &command->op);
	HWFDCAN_set_buffer((uint8_t *) ecommand, sizeof(*command),
			FT_CAN_MASTER_TX_ADDR(txt->common.nslave), FT_CAN_TXN_OP(txt->common.nslave));
}

/**
 * @brief Select type and send new frame
 * @param [in] ecommand command buffer
 * return can do something
 */
int ProcessMasterTxPacket(TXT2_ext_command *ecommand)
{
	for (int i = 1; i <= nslaves; i++) {

		if (HWFDCAN_buffer_busy(FT_CAN_TXN_CONFIG(i)))
			continue;

		struct ft_shm_txt *txt = getTXT(i);

		switch(txt->state) {
		case ftslave_state_addr:
			SendAddrPacket(txt, ecommand);
			break;

		case ftslave_state_ready:
			if (HWFDCAN_buffer_busy(FT_CAN_TXN_OP(i)))
				continue;
			SendConfigPacket(i, ecommand);
			SendOpPacket(txt, ecommand);
			break;

		default:
			continue;
		}
	}

	return 0;
}

